//
//  ENCustomBezelButtonCell.h
//  EverNote
//
//  Created by Matt Sarnoff on 7/9/12.
//  Copyright (c) 2012 Evernote. All rights reserved.
//
//  An NSButtonCell whose bezel is drawn with a custom three-part image.
//
//  To specify the bezel image filename prefix without subclassing, you can add a
//  User Defined Runtime Attribute to the cell in Interface Builder,
//  with a key path of "images" and a type of String.

#import <Cocoa/Cocoa.h>
#import "ENCustomImageCellInternal.h"

@interface ENCustomBezelButtonCell : NSButtonCell
{
  // Pointer to the cell immediately to this cell's right.
  // Should be set if this cell is one of several in an NSMatrix,
  // so that dividers are drawn properly.
  // If this cell is one of several in an NSMatrix acting like a segmented control,
  // the "right" image parts are treated as the dividers.
  IBOutlet ENCustomBezelButtonCell *mNextCell;
  
  ENCustomImageCellInternal *mInternal;
  CGFloat mLeftPadding;
  CGFloat mRightPadding;
  CGFloat mYOffset;
}


/**
 * Sets the bezel image for a specific state and part.
 */
- (void)setImage:(NSImage *)image forState:(ENCustomImageCellState)state part:(ENCustomImageCellPart)part;


/**
 * Sets the three-part bezel image for the normal state.
 */
- (void)setLeftImage:(NSString *)leftImageName middleImage:(NSString *)middleImageName rightImage:(NSString *)rightImageName;


/**
 * Sets the three-part bezel image for the pressed state.
 */
- (void)setPressedLeftImage:(NSString *)leftImageName middleImage:(NSString *)middleImageName rightImage:(NSString *)rightImageName;


/**
 * Set the three-part bezel images for all three states.
 * If namePrefix is "image", this looks for images named:
 *   image-normal-left, image-normal-middle, image-normal-right
 *   image-pressed-left, image-pressed-middle, image-pressed-right
 */
- (void)setImagesWithNamePrefix:(NSString *)namePrefix;


/**
 * KVC-compliant synonym for the above method.
 * This lets you define the images in Interface Builder by adding a
 * User Defined Runtime Attribute named "images" (with type String)
 * to the cell.
 */
- (void)setImages:(NSString *)namePrefix;


/**
 * Or, alternatively, you can override this method in a subclass and return the image name prefix.
 * The images will be loaded in awakeFromNib.
 */
- (NSString *)imageNamePrefix;


/**
 * Prevents setImages: from loading the left or right parts from a set of images.
 */
@property (assign) BOOL noLeftPart;
@property (assign) BOOL noRightPart;


/**
 * Additional interior padding. May be negative.
 */
@property (assign) CGFloat leftPadding;
@property (assign) CGFloat rightPadding;


/**
 * Additional vertical offset for the text. Specify to CGFLOAT_MAX to use standard positioning.
 */
@property (assign) CGFloat yOffset;

@end