// Written by Dirk Holtwick, 2012

/*global Text:false*/
/*global DocumentFragment:false*/

(function() {
  'use strict';
})();

// ========================================================================
// DEBUG
// ========================================================================

if (((typeof window.console) != "undefined") && window['EN_DEBUG']) {
    console.info("ENNoteEditor: Logging is turned on. Level: DEBUG.");

    var log = function() {
        window.console.debug.apply(window.console, arguments);
    };
    var logDebug = function() {
        window.console.debug.apply(window.console, arguments);
    };
    var logInfo = function() {
        window.console.info.apply(window.console, arguments);
    };
    var logWarn = function() {
        window.console.warn.apply(window.console, arguments);
    };
    var logError = function() {
        window.console.error.apply(window.console, arguments);
    };
}

else {
    log =
    logDebug =
    logInfo =
    logWarn =
    logError = function() { };
}

// ========================================================================
// Node
// ========================================================================

Node.prototype.getParent = function(sel, rootNode) {
    if (!rootNode) {
        rootNode = document.body;
    }
    if (this === rootNode) {
        return;
    }
    if (this.matches && this.matches(sel)) {
        return this;
    }
    var p = this.parentElement;
    if (p) {
        return p.getParent(sel);
    }
    return;
};

Node.prototype.isChildOf = function(other) {
    if(!other) {
        return false;
    }
    return !!(this.compareDocumentPosition(other) & 8);
};

Node.prototype.isParentOf = function(other) {
    if(!other) {
        return false;
    }
    return !!(this.compareDocumentPosition(other) & 16);
};

// ========================================================================
// Element
// ========================================================================

if (!Element.prototype.matches) {
    var element = document.createElement('div');
    if (element.webkitMatchesSelector) {
        Element.prototype.matches = function(selector) {
            return this.webkitMatchesSelector(selector);
        };
    }
    else if (element.mozMatchesSelector) {
        Element.prototype.matches = function(selector) {
            return this.mozMatchesSelector(selector);
        };
    }
    else {
        throw 'Unsupported platform';
    }
}

Element.prototype.isParagraph = function() {
    return this.tagName.toLowerCase() == 'p' || this.tagName.toLowerCase() == 'div';
};

Element.prototype.isListContainer = function() {
    return this.tagName.toLowerCase() == 'ul' || this.tagName.toLowerCase() == 'ol';
};

Element.prototype.copyAttributeFromElement = function(el, attrName, defaultValue) {
    var value = el.getAttribute(attrName);
    if (value === null || value === undefined) {
        if (defaultValue === undefined) {
            return null;
        }
        value = defaultValue;
    }
    this.setAttribute(attrName, value);
    return value;
};

// ========================================================================
// NAMESPACE "EN"
// ========================================================================

var EN;
(function (EN) {

// Helper for class inheritance
var __extends = this.__extends || function (d, b) {
    function __() { this.constructor = d; }
    __.prototype = b.prototype;
    d.prototype = new __();
 };

// ========================================================================
// HTMLOPTIMIZER (super class)
// ========================================================================

var HTMLOptimizer = (function () {
    
    // Constructor
    function HTMLOptimizer(inputElement, outputElement, config) {
        var tmpNode = null;
 
        if (typeof config === "undefined") { 
            config = {}; 
        }

        if(config.cssBlockStyle !== undefined) {
            this.cssBlockStyle = config.cssBlockStyle;
            // log(this.cssBlockStyle);
        }
        else {
            this.cssBlockStyle = this.EN_DEFAULT_PARAGRAPH_STYLE;
        }

        this.ios = !!config["ios"];

        if(inputElement instanceof DocumentFragment) {
            tmpNode = document.createElement("div"); 
            tmpNode.appendChild(inputElement);
            document.body.appendChild(tmpNode);
            inputElement = tmpNode;
        }

        if (!inputElement) {
            inputElement = document.body;
        }

        this.inputElement = inputElement;
        this.outputElement = outputElement;

        var resultContainer = this.parseInput();

        if (this.outputElement) {
            this.outputElement.innerHTML = resultContainer.innerHTML;
        }
        else {
            this.outputElement = resultContainer;
        }

        if(tmpNode) {
            document.body.removeChild(tmpNode);
        }
    }

    // Attributes
    HTMLOptimizer.prototype.EN_DEFAULT_PARAGRAPH_TAG = 'div';
    HTMLOptimizer.prototype.EN_DEFAULT_PARAGRAPH_STYLE = null;
    HTMLOptimizer.prototype.EN_EMPTY_LINE_THRESHOLD_VALUE = 0.8;
    HTMLOptimizer.prototype.EN_DEFAULT_PARAGRAPH_STYLE = null;

    HTMLOptimizer.prototype.parse = function () {
        this.source = null;
    };

    HTMLOptimizer.prototype.parseInput = function() {

        // We build the simplified text into a new structure
        this.block = null;
        this.blockReferenceNode = null;
        this.listReferenceNode = null;
        this.resultElement = document.createElement('div');
        this.lastListItem = null;
        this.lastBBox = null;
        this.requiresEmptyLine = false;

        // Threshold for empty lines is 80% of default font size
        this.thresholdForEmptyLine = window.getComputedStyle(document.body).fontSize.replace(/px/gi, '') * this.EN_EMPTY_LINE_THRESHOLD_VALUE;
        // logDebug("Threshold in pixel for empty line is", this.thresholdForEmptyLine);

        // Create tree walker, collpase entities to text nodes
        var treeWalker = document.createTreeWalker(this.inputElement,
                NodeFilter.SHOW_ALL, {
                    acceptNode: function(node) {
                        return NodeFilter.FILTER_ACCEPT;
                    }
                }, true);

        // Iterate through the tree and create something more simple
        while (treeWalker.nextNode()) {
            var node = treeWalker.currentNode;
            if (node.nodeType == Node.TEXT_NODE) {
                this.handleText(node);
            }
            else if (node.nodeType == Node.ELEMENT_NODE) {
                this.handleElement(node);
            }
        }

        // The last block has to be touched again too
        this.optimizeFinishedBlockElement(this.block);

        return this.resultElement;
    };

    HTMLOptimizer.prototype.stripTrailingBlanks = function(blockElement) {
        var n;    
        var iter= document.createNodeIterator(
              blockElement, 
              NodeFilter.SHOW_TEXT,
              null,
              true);
        
        // Strip unneeded from start
        while ((n = iter.nextNode())) {
            n.textContent = n.textContent.replace(/^ /gi, "");        
            if (n.textContent.length > 0) {
                break;
            }
        }

        // Move to end
        while (iter.nextNode()) { }

        // Iterate from end to start
        while ((n = iter.previousNode())) {
            n.textContent = n.textContent.replace(/\s+$/gi, "");        
            if(n.textContent.length > 0) {
                break;
            }
        }
    };

    HTMLOptimizer.prototype.optimizeFinishedBlockElement = function(blockElement) {
        // No element, no work
        if(!blockElement) return;    
        // Remove trailing white space
        this.stripTrailingBlanks(blockElement);    
        // Combine fragmented text nodes
        blockElement.normalize();
    };

    HTMLOptimizer.prototype.createBlockElement = function(tag) {
        var el = document.createElement(!!tag ? tag : this.EN_DEFAULT_PARAGRAPH_TAG);
        if(this.cssBlockStyle) {
           el.setAttribute("style", this.cssBlockStyle);
        }
        return el;
    };

    HTMLOptimizer.prototype.appendEmptyLine = function() {
        var br = this.createBlockElement();
        br.appendChild(document.createElement('br'));
        this.resultElement.appendChild(br);
        this.requiresEmptyLine = false;
    };

    HTMLOptimizer.prototype.appendBlockElement = function(tag) {
        var el = this.createBlockElement(tag);

        // Do we need to add an additional blank line?
        if (this.resultElement.hasChildNodes && this.requiresEmptyLine) {
            var last = this.resultElement.lastElementChild;

            // Only continue if
            if (last && last.textContent.length && last.isParagraph() && el.isParagraph()) {
                this.appendEmptyLine();
            }

        }
        this.requiresEmptyLine = false;

        // Apend block
        this.resultElement.appendChild(el);
        return el;
    };

    HTMLOptimizer.prototype.startNewBlock = function(keepBlockReference) {

        // Optimize the last finished block
        this.optimizeFinishedBlockElement(this.block);

        // Reset to trigger new block when we encounter a text portion
        this.block = null;

        if (!keepBlockReference) {
            this.blockReferenceNode = null;
        }
    };

    HTMLOptimizer.prototype.calculateExtraBlankLines = function(node) {
        // Calculate distances
        // FIXME:dholtwick:2012-09-19 - Absolute or floating elements should be considered far away
        // var bbox = node.parentElement.getBoundingClientRect();
        var range = document.createRange();
        range.selectNodeContents(node);
        var bbox = range.getBoundingClientRect();

        if (this.lastBBox && bbox && !node.isChildOf(this.blockReferenceNode)) {
            var deltaY = Math.abs(bbox.top - this.lastBBox.bottom);        
            // log(deltaY, bbox, node.parentElement, window.getComputedStyle(node));

            if ((deltaY > this.thresholdForEmptyLine)) {
                // Trigger empty line
                this.requiresEmptyLine = true;
            }
        }
        if (bbox) {
            this.lastBBox = bbox;
        }
    };

    HTMLOptimizer.prototype.appendInlineElement = function(node, elementToInsert, onTopLevel) {
        var tag;
                     
        if (!elementToInsert) {
            elementToInsert = node.cloneNode(false);
        }

        this.calculateExtraBlankLines(node);

        // Did we leave the context of the last block?
        if (!node.isChildOf(this.blockReferenceNode)) { 

            // Some elements need special treatment
            if(this.blockReferenceNode instanceof Element) {
                tag = this.blockReferenceNode.tagName.toLowerCase();
                if (tag == 'td') {
                    this.requiresEmptyLine = true;
                }
            }

            this.startNewBlock();
        }

        // Find block element
        var container = node;
        while ((container != this.inputElement) && (container = container.parentNode)) {
            this.blockReferenceNode = container;
            tag = container.tagName.toLowerCase();
            if (tag == 'li') {
                break;
            }
            if (tag == 'td') {
                break;
            }
            if (window.getComputedStyle(container).display.toLowerCase() == 'block') {
                break;
            }
        }

        // Inside list? This is something special for us
        var nodeListItem = node.getParent('li');
        if (!this.block && nodeListItem) {
            var listElement = null;

            // Whitespace in lists is ignored and should collapse
            if (node instanceof Text) {
                if (node.textContent.search(/^\s*$/i) !== -1) {
                    return;
                }
                elementToInsert = document.createTextNode(node.textContent.replace(/\s+/gi, ' '));
            }

            if (this.lastListItem && this.listReferenceNode == nodeListItem) { 
                this.block = this.lastListItem;
                this.block.appendChild(document.createElement('br'));
            }

            else { 

                // Analyze list structure
                var listLevel = 0;
                var listType = null;

                var cursor = nodeListItem;
                while (cursor && (cursor !== this.inputElement)) {
                    if (cursor.isListContainer()) {

                        // We count these, because often lists are not nested correctly
                        // i.e. they are not inside a <li>
                        ++listLevel;

                        // This is the type of the list element (ol/ul)
                        if (!listType) {
                            listType = cursor.tagName.toLowerCase();
                        }
                    }
                    cursor = cursor.parentElement;
                }

                listLevel = Math.max(0, listLevel); 

                // Get the matching list container
                var lastListContainer = null;
                cursor = this.resultElement;
                while (1) {
                    cursor = cursor.lastElementChild;
                    if (!cursor) {
                        break;
                    }
                    if (cursor.isListContainer()) {
                        lastListContainer = cursor;
                        --listLevel;
                        if (listLevel < 1) {
                            break;
                        }
                    }
                }

                // Previous element was not a list or of same type
                if (listLevel > 0 && lastListContainer) {
                    var li = lastListContainer.lastElementChild;
                    lastListContainer = this.appendBlockElement(listType || 'ol');
                    li.appendChild(lastListContainer);
                }

                if (!lastListContainer) {
                    lastListContainer = this.appendBlockElement(listType || 'ol');
                }

                this.block = this.createBlockElement('li');
                lastListContainer.appendChild(this.block);

                this.listReferenceNode = nodeListItem;
                this.lastListItem = this.block;
            }
        }
        else {

            // Still no block, then make a paragraph
            this.listReferenceNode = null;
            if (!this.block) {
                this.block = this.appendBlockElement();
            }
        }

        // Collect inline style flags
        var flags = this.getStyleFlagsOfNode(node);

        // Add text to block
        // FIXME:dholtwick:2012-09-16 - Adds a lot of noise, simplify e.g. by merging equal siblings...
        this.block.appendChild(
            this.sourroundNodeWithStyles(
                elementToInsert,
                flags));

        return elementToInsert;
    };

    HTMLOptimizer.prototype.handleText = function(node) {
        var txt;
                     
        // PRE?
        if ((/pre/gi).test(window.getComputedStyle(node.parentElement).whiteSpace)) {

            // Split into lines
            var lines = node.textContent.split(/\n\r|\r\n|\r|\n/gi);
     
            if (lines && lines.length > 0) {
                // Preserve &nbsp; but collapse the other whitespaces
                txt = lines[0].replace(/\s/gi, ' ').replace(/\s\s/gi, '\xa0 ');

                var txtNode = document.createTextNode(txt);
                this.appendInlineElement(node, txtNode);

                for (var i = 1; i < lines.length; i++) {

                    txtNode.parentNode.appendChild(document.createElement('br'));

                    txt = lines[i].replace(/\s/gi, ' ').replace(/\s\s/gi, '\xa0 ');
                    txtNode.parentNode.appendChild(document.createTextNode(txt));
                }
            }
        }
        else {

            // Preserve &nbsp; but collapse the other whitespaces
            txt = node.textContent.replace(/[\ \n\r\t]+/gi, ' ');

            // If empty ignore
            if (txt.search(/^[\ \t\n]*$/i) !== -1) {
                return;
            }

            this.appendInlineElement(node, document.createTextNode(txt));
        }
    };

    HTMLOptimizer.prototype.handleElement = function(node) { 
        var tag = node.tagName.toLowerCase();

        // HR are encapsualted in DIV
        if (tag == 'hr') {
            this.startNewBlock();
            this.appendInlineElement(node, document.createElement('hr'), true);
            this.startNewBlock();
            return;
        }

        // Images get a new line
        if (tag == 'img' && node.src) { 
            this.startNewBlock();
            var img = document.createElement('img');
            img.copyAttributeFromElement(node, 'src');
            img.copyAttributeFromElement(node, 'alt', '');
            img.copyAttributeFromElement(node, 'id');
            if(this.ios) {
                img.copyAttributeFromElement(node, 'class');
                img.copyAttributeFromElement(node, 'x-evernote-mime');                
            }
            this.appendInlineElement(node, img, true);
            this.startNewBlock();
            return;
        }
     
        // Mac embedded objects
        if (!this.ios && tag == 'embed') { 
            var objectType = node.getAttribute('type').toLowerCase();
            if (objectType == 'evernote/x-todo' || objectType == 'evernote/x-attachment' || objectType == 'evernote/x-pdf') {
                this.appendInlineElement(node);
            } 
            return;
        }

        // iOS embedded objects
        if(this.ios && tag == "object") {
            this.appendInlineElement(node);
            return;
        }

        if(this.ios && tag == "input") {
            var klass = node.getAttribute('class').toLowerCase();
            if(klass == "en-todo") {
                this.appendInlineElement(node);
                return;
            }
        }

        // Handle blank lines
        if (tag == 'br') {
            this.startNewBlock();
            // this.appendEmptyLine();
            return;
        }

        // Handle table cells
        if (tag == 'td') {               
            this.startNewBlock();
            this.requiresEmptyLine = true;
            return;
        }

        if (this.block && (
            // Case #1 Per style definition this behaves as a block element
            (window.getComputedStyle(node).display.toLowerCase() == 'block') ||
            // Case #2: We left the context of out former block defining element
            (!this.blockReferenceNode.isParentOf(node))))
        {
            this.startNewBlock();
            return;
        }
    };

    HTMLOptimizer.prototype.getStyleFlagsOfNode = function(node) {

        var flags = {};
        while (node || node == this.inputElement) {
            var tag = node.tagName && node.tagName.toLowerCase();
            // if (tag == 'b' || tag == 'strong') {
            //  flags.b = 1;
            // }
            // else if (tag == 'i' || tag == 'em') {
            //  flags.i = 1;
            // }
            // else if (tag == 'u') {
            //  flags.u = 1;
            // }
            if (tag == 'del' || tag == 'strike' || tag == 's') {
                flags.s = 1;
            }
            else if (tag == 'mark') {
                flags.m = 1;
            }
            else if (tag == 'a') {
                flags.a = node.href;
            }

            // Other inline element like SPAN or FONT might also contain style info
            var style = window.getComputedStyle(node);
            if (style /* && style.display == 'inline' */) {
                if ( // From iOS
                    style.backgroundColor.toLowerCase() == 'rgb(255, 204, 102)' ||
                    // From Clearly
                    style.backgroundColor.toLowerCase() == 'rgb(246, 238, 150)')                 
                {
                    flags.m = 1;
                }

                var fontWeight = style.fontWeight.toLowerCase();
                var fontStyle = style.fontStyle.toLowerCase();
                var textDecorations = style.textDecoration.toLowerCase();

                if (fontWeight == 'bold' || fontWeight == 'bolder') {
                    flags.b = 1;
                }

                if (fontStyle == "italic" || fontStyle == "oblique") {
                    flags.i = 1;
                }

                // Inside links it is natural that we have undelined text
                if ((!flags['a']) && (textDecorations.indexOf('underline') >= 0)) {
                    flags.u = 1;
                }

                if (textDecorations.indexOf('line-through') >= 0) {
                    flags.s = 1;
                }
            }
            node = node.parentNode;
        }

        // Links are often underlined, we strip it
        if (flags['a'] && flags['u']) {
            delete flags.u;
        }

        return flags;
    };

    HTMLOptimizer.prototype.sourroundNodeWithStyles = function(node, flags) {
        var el;
        if (flags.a) {
            el = document.createElement('a');
            el.setAttribute('href', flags.a);
            el.appendChild(node);
            node = el;
        }
        if (flags.b) {
            el = document.createElement('strong');
            el.appendChild(node);
            node = el;
        }
        if (flags.i) {
            el = document.createElement('em');
            el.appendChild(node);
            node = el;
        }
        if (flags.u) {
            el = document.createElement('u');
            el.appendChild(node);
            node = el;
        }
        if (flags.s) {
            el = document.createElement('del');
            el.appendChild(node);
            node = el;
        }
        if (flags.m) {
            el = document.createElement('mark');
            el.appendChild(node);
            node = el;
        }
        return node;
    };

    return HTMLOptimizer;
})();    

// ========================================================================
// SIMPLE TEXT (extends HTMLOptimizer)
// ========================================================================

var SimpleText = (function (_super) {
    function SimpleText() {
        _super.apply(this, arguments);
    }
    __extends(SimpleText, _super);                  
    return SimpleText;
})(HTMLOptimizer);

function createNewSimpleTextElement(inputElement) {
    var obj = new SimpleText(inputElement, null);
    return obj.outputElement;
}

// ========================================================================
// PLAIN TEXT (extends HTMLOptimizer)
// ========================================================================

var PlainText = (function (_super) {
    function PlainText() {
        _super.apply(this, arguments);
    }
    __extends(PlainText, _super);

    // Attributes
    PlainText.prototype.EN_DEFAULT_PARAGRAPH_TAG = 'div';
    PlainText.prototype.EN_DEFAULT_PARAGRAPH_STYLE = 'font-family: Courier, "Courier New", monospace;';
    PlainText.prototype.EN_KEEP_LINKS = false;
    PlainText.prototype.EN_LIST_INDENT = "\xa0 \xa0 ";
    PlainText.prototype.EN_LIST_BULLET = "- ";
    PlainText.prototype.EN_LIST_INDENT_FIRST = false;
 
    PlainText.prototype.parseInput = function() {

        // We remember attachments and append them to the end of the document
        this.attachments = [];
        
        _super.prototype.parseInput.call(this);

        // Append attachments
        this.requiresEmptyLine = true;
        for (var i = 0; i < this.attachments.length; i++) {
            this.block = this.appendBlockElement();
            this.block.appendChild(this.attachments[i]);
        }

        return this.resultElement;
    };


    // No exceptions, only plain divs
    PlainText.prototype.createBlockElement = function(tag) {
        var el = document.createElement(this.EN_DEFAULT_PARAGRAPH_TAG);
        if(this.cssBlockStyle) {
            el.setAttribute("style", this.cssBlockStyle);
        }
        return el;
    };

    PlainText.prototype.appendInlineElement = function(node, elementToInsert) {
        var tag;

        if (!elementToInsert) {
            elementToInsert = node.cloneNode(false);
        }

        this.calculateExtraBlankLines(node);

        // Did we leave the context of the last block?
        if (!node.isChildOf(this.blockReferenceNode)) { 

            // Some elements need special treatment
            if(this.blockReferenceNode instanceof Element) {
                tag = this.blockReferenceNode.tagName.toLowerCase();
                if (tag == 'td') {
                    this.requiresEmptyLine = true;
                }
            }

            this.startNewBlock();
        }

        // Find block element
        var container = node;
        while ((container != this.inputElement) && (container = container.parentNode)) {
            this.blockReferenceNode = container;
            tag = container.tagName.toLowerCase();
            if (tag == 'li') {
                break;
            }
            if (tag == 'td') {
                break;
            }
            if (window.getComputedStyle(container).display.toLowerCase() == 'block') {
                break;
            }
        }

        // Inside list? This is something special for us
        var nodeListItem = node.getParent('li');
        if (!this.block && nodeListItem) {
            var listElement = null;

            // Whitespace in lists is ignored and should collapse
            // if (node instanceof Text) {
            //     if (node.textContent.search(/^\s*$/i) == 0) {
            //         return;
            //     }
            //     elementToInsert = document.createTextNode(node.textContent.replace(/\s+/gi, ' '));
            // }

            if (this.lastListItem && this.listReferenceNode == nodeListItem) { 
                this.block = this.lastListItem;
                this.block.appendChild(document.createElement('br'));
            }

            else { 

                // Analyze list structure
                var listLevel = 0;
                var cursor = nodeListItem;
                while (cursor && (cursor !== this.inputElement)) {
                    if (cursor.isListContainer()) {
                        ++listLevel;
                    }
                    cursor = cursor.parentElement;
                }
                listLevel = Math.max(0, listLevel); 
 
                this.block = this.appendBlockElement();

                // List bullet and indent
                var prefix = this.EN_LIST_BULLET;
                while(listLevel > 1) {
                    prefix = this.EN_LIST_INDENT  + prefix;
                    --listLevel;
                }
                this.block.appendChild(document.createTextNode(prefix));

                this.listReferenceNode = nodeListItem;
                this.lastListItem = this.block;
            }
        }
        else {

            // Still no block, then make a paragraph
            this.listReferenceNode = null;
            if (!this.block) {
                this.block = this.appendBlockElement();
            }
        }

        // Collect inline style flags
        var flags = this.getStyleFlagsOfNode(node);

        // Add text to block
        // FIXME:dholtwick:2012-09-16 - Adds a lot of noise, simplify e.g. by merging equal siblings...
        this.block.appendChild(
            this.sourroundNodeWithStyles(
                elementToInsert,
                flags));

        return elementToInsert;
    };

    PlainText.prototype.handleElement = function(node) { 
        var tag = node.tagName.toLowerCase();

        if (tag == 'hr') {
            this.startNewBlock();
            return;
        }

        // Images get a new line
        if (tag == 'img') { 
            return;
        }
     
        if (!this.ios && tag == 'embed') { 
            var objectType = node.getAttribute('type').toLowerCase();
            if (objectType == 'evernote/x-todo') {
                var checked = (node.getAttribute("todo-checked") + "").toLowerCase() == "true";
                this.appendInlineElement(node, document.createTextNode(checked ? "[x] " : "[ ] "));
            } 
            else { //if(objectType == 'evernote/x-attachment' || objectType == 'evernote/x-pdf') {
                this.attachments.push(node.cloneNode(true));
            }
            return;
        }

        if(this.ios && tag == "input") {
            var klass = node.getAttribute('class').toLowerCase();
            if(klass == "en-todo") {
                this.appendInlineElement(node, document.createTextNode(node.checked ? "[x] " : "[ ] "));
                return;
            }
        }

        if(this.ios && tag == "object") {
            this.appendInlineElement(node);
            return;
        }

        // Handle blank lines
        if (tag == 'br') {
            this.startNewBlock();
            return;
        }

        // Handle table cells
        if (tag == 'td') {               
            this.startNewBlock();
            this.requiresEmptyLine = true;
            return;
        }

        if (this.block && (
            // Case #1 Per style definition this behaves as a block element
            (window.getComputedStyle(node).display.toLowerCase() == 'block') ||
            // Case #2: We left the context of out former block defining element
            (!this.blockReferenceNode.isParentOf(node))))
        {
            this.startNewBlock();
            return;
        }
    };

    PlainText.prototype.getStyleFlagsOfNode = function(node) {
        var flags = {};
        if (this.EN_KEEP_LINKS) {
            while (node || node == this.inputElement) {
                var tag = node.tagName && node.tagName.toLowerCase();
                if (tag == 'a') {
                    flags.a = node.href;
                }
                node = node.parentNode;
            }

            // Links are often underlined, we strip it
            if (flags['a'] && flags['u']) {
                delete flags.u;
            }
        }
        return flags;
    };

    PlainText.prototype.sourroundNodeWithStyles = function(node, flags) {
        if (this.EN_KEEP_LINKS) {
            if (flags.a) {
                var el = document.createElement('a');
                el.setAttribute('href', flags.a);
                el.appendChild(node);
                node = el;
            }
        }
        return node;
    };

    return PlainText;
})(HTMLOptimizer);

function createNewPlainTextElement(inputElement, cssBlockStyle) {
    var config = {};
    if(cssBlockStyle) {
        config.cssBlockStyle = cssBlockStyle;
    }
    var obj = new PlainText(inputElement, null, config);
    return obj.outputElement;
}

// ========================================================================
// MAC SPECIFIC
// ========================================================================

function computedStyleOfCurrentSelection() {
    var sel = window.getSelection();
    if (sel.rangeCount < 1) {
        return null;
    }

    var el = sel.getRangeAt(0).startContainer;
    var styles;

    while (!(styles = window.getComputedStyle(el)) && el) {
        el = el.parentNode;
        if (!el) {
            styles = null;
            break;
        }
    }

    if (!el || !styles) {
        return null;
    }

    return styles;
}

function fontFamilyAtSelection() {
    var styles = computedStyleOfCurrentSelection();

    if (!styles || !styles['font-family']) {
        return "";
    }

    return (styles['font-family'].replace(/^.*?([a-z0-9\-\ ]+).*$/gi, '$1')).toString();
}

function fontSizeInPixelAtSelection () {
    var styles = computedStyleOfCurrentSelection();

    if (!styles || !styles['font-size']) {
        return "";
    }

    return (styles['font-size'].replace(/px/gi, '')).toString();
} 

// ========================================================================
// SELECTION HELPER
// ========================================================================

// Range/ Selection helper
// ------------------------------------------------------------------------

function selectionWithRange(range) {
    var sel = window.getSelection();
    sel.removeAllRanges();
    sel.addRange(range);
    return sel;
}

function rangeFromSelection() {
    var sel = window.getSelection();
    if (sel.rangeCount < 1) {        
        return null;
    }
    return sel.getRangeAt(0);
}

// Node and range descriptions as offset path
// ------------------------------------------------------------------------

function getNodeLength(node) {
  // Idea borrowed from http://closure-library.googlecode.com/svn/docs/closure_goog_editor_range.js.source.html
  return node.length || node.childNodes.length;
}

function nodeWithOffsetPath(root, offsetPath) {
    if(!root) {
        root = document.body;
    } 
    var node = root;  
    for (var i = offsetPath.length - 1; i >= 0; i--) {
        if(getNodeLength(node) <= 0) {
            logWarn("Node is empty", node);
        }
        var offset = offsetPath[i];        
        if(offset > getNodeLength(node)) {
            logWarn("Offset to high", offsetPath[i], node);
            offset = getNodeLength(node) - 1;
        }
        node = node.childNodes[offset]; 
    }    
    return node;
}

function getNodeOffsetInParent(theNode) {
    var offset = 0;
    var children = theNode.parentNode.childNodes;
    for (var i = 0; i < children.length; i++) {
        if (children[i] === theNode) {
            offset = i;
            break;
        }
    } 
    return offset;    
}

function getNodeOffsetInParentForNormalizedDOM(theNode) {
    var offset = 0, textOffset = 0;
    var children = theNode.parentNode.childNodes;
    for (var i = 0; i < children.length; i++) {
        var node = children[i];
        if(node.nodeType !== Node.TEXT_NODE) {
            if (textOffset > 0) {
                ++offset;
                textOffset = 0;   
            }             
        }                
        if (node === theNode) {
            break;
        }
        if(node.nodeType === Node.TEXT_NODE) {
            textOffset += getNodeLength(node);
        }                
        else {
            ++offset;
        }
    } 
    return offset;    
}

Node.prototype.offsetPath = function(root) {
    if (!root) {
        root = document.body;
    }  

    var nodePath = [];
    var node = this;
    while (node && node.parentNode && (node !== root)) {
        nodePath.push(getNodeOffsetInParent(node));
        node = node.parentNode;
    }
    return nodePath;
};

function getDeepestPointForNodeAndOffset(node, offset, opt_trendLeft) {
    // Idea borrowed from http://closure-library.googlecode.com/svn/docs/closure_goog_editor_range.js.source.html
    while (node.nodeType == Node.ELEMENT_NODE) {
        var child = node.childNodes[offset];
        if (!child && !node.lastChild) {
            break;
        }
        if (child) {
            var prevSibling = child.previousSibling;
            if (opt_trendLeft && prevSibling) {
                node = prevSibling;
                offset = getNodeLength(node);
            } 
            else {
                node = child;
                offset = 0;
            }
        } 
        else {
            node = node.lastChild;
            offset = getNodeLength(node);
        }
    }
    return {
        node: node, 
        offset: offset
    };
}

function getPointForNormalizedDOM(point) {
    // Idea borrowed from http://closure-library.googlecode.com/svn/docs/closure_goog_editor_range.js.source.html
    var previous;
    if (point.node.nodeType == Node.TEXT_NODE) {
        for (var current = point.node.previousSibling;
             current && current.nodeType == Node.TEXT_NODE;
             current = current.previousSibling) 
        {
            point.offset += getNodeLength(current);
        }
        previous = current;
    } 
    else {
        previous = point.node.previousSibling;
    }

    var parent = point.node.parentNode;
    point.node = previous ? previous.nextSibling : parent.firstChild;
    return point;
}

function getPointSnapshotForNormalizedDOM(root, node, offset) {

    var point;
    point = getDeepestPointForNodeAndOffset(node, offset);
    point = getPointForNormalizedDOM(point);
    node = point.node;             

    var nodePath = [];        
    while (node && node.parentNode && (node !== root)) {
        nodePath.push(getNodeOffsetInParentForNormalizedDOM(node));
        node = node.parentNode;
    }

    return {
        path: nodePath, 
        offset: point.offset
    };
}

Range.prototype.setFromOffsetPath = function(root, path) {
    if(!root) {
        root = document.body;
    }
    if (!root || !path) {
        throw 'All arguments required';
    }

    this.setStart(nodeWithOffsetPath(root, path.startPath), path.startOffset);
    this.setEnd(nodeWithOffsetPath(root, path.endPath), path.endOffset);

    if(path.collapsed) {
        this.collapse(true);
    }

    return this;
};

// Cursor and selection public methods
// ------------------------------------------------------------------------

/*  A selection snapshot is an object returned as a JSON string which
    contains all values needed to restore a selection. This is done
    by transforming the selections node to a list of recursive offsets 
    of the node within its parent. The list starts at `root`, which by
    default is `document.body`.

    Something special is the option `forNormalizedDOM`. After restoring 
    HTML e.g. by using `innerHTML` or by calling `root.normailze()` all
    adjacent text nodes collapse to one or are deleted if empty. This 
    changes the offset path described before, therefore this option
    allows to calculate the path as if the DOM were normailzed.

    Of course it would be much simpler to first apply normalization and
    then call this routine, but selection usually get lost or are not
    correct any more in many browsers. Also we need to keep the original
    DOM for the UNDO in Mac client to keep working.
*/
function getSelectionSnapshot(root, forNormalizedDOM) {
    if(!root) {
        root = document.body;
    }
    var selectionPath = "";
    var selectionRange = rangeFromSelection();
    if(selectionRange) {
        if(!forNormalizedDOM) {
            selectionPath = {
                startPath: selectionRange.startContainer.offsetPath(root),
                startOffset: selectionRange.startOffset,
                endPath: selectionRange.endContainer.offsetPath(root),
                endOffset: selectionRange.endOffset,
                collapsed: selectionRange.collapsed
            }; 
        }
        else {
            var startSnap = getPointSnapshotForNormalizedDOM(root, selectionRange.startContainer, selectionRange.startOffset);
            var endSnap = getPointSnapshotForNormalizedDOM(root, selectionRange.endContainer, selectionRange.endOffset);
            selectionPath = {
                startPath: startSnap.path,
                startOffset: startSnap.offset,
                endPath: endSnap.path,
                endOffset: endSnap.offset,
                collapsed: this.collapsed
            };
        }
    }
    return JSON.stringify(selectionPath);
} 

function restoreSelectionSnapshot(root, selectionSnapshot) {
    if(!root) {
        root = document.body;
    }

    // Get snapshot of current selection
    var currentSnapshot = getSelectionSnapshot(root);

    // Restore
    var snapshot = JSON.parse(selectionSnapshot);
    var range = document.createRange();
    range.setFromOffsetPath(root, snapshot);
    selectionWithRange(range);

    return currentSnapshot;
}

// ========================================================================
// UNDO
// ========================================================================

/*  The puropse of this function is to remember the current
    selection and clone the children of root. Then the children
    of root are replaced by the clones and the selection will
    be restored. This is needed in order to make UNDO on 
    Mac to function correctly.
*/
function prepareUNDO(root) {

    if(!root) {
        root = document.body;
    }

    // Remember selection
    var selSnapshot = getSelectionSnapshot(root);

    // Create a clone of root
    // We need to exactly do it this way to avoid the reuse
    // of any nodes, because Cocoa tracks them for UNDO
    var rootRange = document.createRange();
    rootRange.selectNodeContents(root);
    var clonedContents = rootRange.cloneContents();

    // Remove old content
    while(root.firstChild) {
        root.removeChild(root.firstChild);
    }

    // Re insert content
    root.appendChild(clonedContents);

    // Restore selection
    return restoreSelectionSnapshot(root, selSnapshot);
}

// ========================================================================
// TEXT TRANSFORMATIONS
// ========================================================================

function applyTextTransformations(root, transformationFunc) {
    var newNode;
    if(!root) {
        root = document.body;
    }

    // Get current selection
    var range = rangeFromSelection();
    if(!range || range.toString().length < 1) {
        return;
    }

    // Split the text nodes if needed
    if(range.endContainer instanceof Text && range.endOffset > 0 && range.endOffset < range.endContainer.textContent.length) {
        newNode = range.endContainer.splitText(range.endOffset);        
        range.setEnd(newNode, 0);
    }
    if(range.startContainer instanceof Text && range.startOffset > 0) {
        newNode = range.startContainer.splitText(range.startOffset);        
        range.setStart(newNode, 0);
    }
     
    // We need to set the selection again, because we use it to determine 
    // which nodes are inside the selection
    selectionWithRange(range);

    // Remember the frst and last text node to adjust the selection afterwards
    var startTextNode = null;
    var endTextNode = null;

    // Get all text nodes inside the selection
    var sel = window.getSelection();
    var treeWalker = document.createTreeWalker(
        root,
        NodeFilter.SHOW_TEXT,
        function(node) {
            return sel.containsNode(node, false) ? NodeFilter.FILTER_ACCEPT : NodeFilter.FILTER_REJECT;
        },
        false
    );
    while (treeWalker.nextNode()) {
        var textNode = treeWalker.currentNode;        
        if(textNode instanceof Text) {
            treeWalker.currentNode.textContent = transformationFunc(treeWalker.currentNode.textContent);
            if(!startTextNode) {
                startTextNode = treeWalker.currentNode;
            }
            endTextNode = treeWalker.currentNode;
        }
    }

    // Update the selection, it got lost while manipulating the text nodes
    range = document.createRange();
    range.setStartBefore(startTextNode);
    range.setEndAfter(endTextNode);
    selectionWithRange(range);
}

function applyUpperCase(root) {
    if(!root) {
        root = document.body;
    }
    applyTextTransformations(root, function(text) {
        return text.toLocaleUpperCase();
    });
}

function applyLowerCase(root) {
    if(!root) {
        root = document.body;
    }
    applyTextTransformations(root, function(text) {
        return text.toLocaleLowerCase();
    });
}

function normalizeSelectionSafe(root) {
    var selSnap = getSelectionSnapshot(root, true);
    root.normalize();
    restoreSelectionSnapshot(root, selSnap);
    return root;
}

function applyCapitalize(root) {
    if(!root) {
        root = document.body;
    }

    normalizeSelectionSafe(root);    

    // FIXME:dholtwick:2012-10-16 
    // May fail with non latin languages
    applyTextTransformations(root, function(text) {
        var newText = "";        
        var words = text.split(/(\W+)/gi);
        for (var i = 0; i < words.length; i++) {
            var word = words[i];            
            if(word.length > 0) {
                newText += word.charAt(0).toLocaleUpperCase() + word.slice(1).toLocaleLowerCase();
            }
            else {
                newText += word;
            }
        }
        return newText;
    });
}

// ========================================================================
// EXECCOMMAND / UNDO COMPATIBLE 
// ========================================================================

function resetDocumentUsingExecCommand() {
    // This might look lige overkill, but this is all needed to get rid of
    // all formattings, list nestings and inline styles. Believe me ;)
    document.execCommand("selectAll", false, null);
    document.execCommand("delete", false, null);    
    document.execCommand("selectAll", false, null);
    document.execCommand("formatBlock", false, "div");
    document.execCommand("delete", false, null);    
    document.execCommand("selectAll", false, null);
}

function simpleTextUsingExecCommand() {
    var obj = new SimpleText(document.body, null, {
        ios: true
    });
    var html = obj.outputElement.innerHTML; 
    resetDocumentUsingExecCommand();
    document.execCommand("insertHTML", false, html);
}

function plainTextUsingExecCommand() {
    var obj = new PlainText(document.body, null, {
        ios: true
    });
    var html = obj.outputElement.innerHTML;
    resetDocumentUsingExecCommand();
    document.execCommand("insertHTML", false, html);
}

// ========================================================================
// PUBLIC 
// ========================================================================

EN.SimpleText = SimpleText;
EN.PlainText = PlainText;

EN.createNewSimpleTextElement = createNewSimpleTextElement;
EN.createNewPlainTextElement = createNewPlainTextElement;

EN.fontFamilyAtSelection = fontFamilyAtSelection;
EN.fontSizeInPixelAtSelection = fontSizeInPixelAtSelection;

EN.rangeFromSelection = rangeFromSelection;
EN.selectionWithRange = selectionWithRange;

EN.restoreSelectionSnapshot = restoreSelectionSnapshot;
EN.getSelectionSnapshot = getSelectionSnapshot;

EN.prepareUNDO = prepareUNDO;

EN.applyUpperCase = applyUpperCase;
EN.applyLowerCase = applyLowerCase;
EN.applyCapitalize = applyCapitalize;

EN.simpleTextUsingExecCommand = simpleTextUsingExecCommand;
EN.plainTextUsingExecCommand = plainTextUsingExecCommand;

// End of namespace EN
})(EN || (EN = {}));

